function [y,nz] = ompdenoise(params,msgdelta)
%OMPDENOISE OMP denoising.
%  Y = OMPDENOISE(PARAMS) denoises a (possibly multi-dimensional) signal
%  using OMP denoising. OMPDENOISE denoises each block of the noisy signal
%  individually by first removing its mean and then performing OMP. The
%  denoised blocks are averaged when they overlap in the output signal.
%
%  Y = OMPDENOISE(PARAMS,MSGDELTA) specifies whether the function should
%  print messages as it runs. When MSGDELTA is 0, no messages are
%  displayed. When MSGDELTA is larger than zero, a message is displayed
%  approximately every MSGDELTA seconds.
%
%  [Y,NZ] = KSVDDENOISE(...) also returns the average number of non-zero
%  coefficients in the representations of the denoised blocks.
%
%
%  Required fields in PARAMS:
%    x          -  the signal to denoise (can be multi-dimensional). x
%                  should be of type double, and (for psnr computations)
%                  its values are assumed to be in the range [0,1] (see
%                  parameter maxval below).
%    blocksize  -  the block size for the denoising. should be either an
%                  array of the form [N1 N2 ... Np], where p is the
%                  number of dimensions of x, or simply a scalar N, which
%                  is shorthand for the square block size [N N ... N].
%    D          -  the dictionary used to denoise each block in x
%    psnr/sigma -  noise power (in dB) or noise standard deviation. used to
%                  determine the target error for sparse coding each block.
%                  if both are present, sigma is used unless the field
%                  'noisemode' is specified (see below). when specifying
%                  the noise power in psnr, make sure to set the 'maxval'
%                  parameter (below) if needed.
%
%
%  Optional fields in PARAMS:
%    maxval     -  the maximal intensity value allowed in the signal. used
%                  to determine the noise standard deviation when the noise
%                  power is specified in PSNR. the default computation
%                  assumes signals with samples in the range [0,1]. if the
%                  maxval parameter is specified, the range is assumed to
%                  be [0,maxval].
%    stepsize   -  the interval in voxels between neighboring signal
%                  blocks. by default, all overlapping signal blocks are
%                  denoised and averaged. when stepsize is set to the array
%                  [S1 S2 ... Sp] (where p is the number of dimensions of
%                  x), the distance between processed blocks in the i-th
%                  dimension becomes Si. stepsize can also be specified as
%                  a scalar S, shorthand for the step size [S S ... S].
%                  Each Si must be >= 1, and, to avoid cases where the
%                  boundary of x is not covered by any block, it is
%                  recommended that (size(X,i)-Ni) be a multiple of Si. the
%                  default behavior of the function corresponds to the
%                  setting stepsize==1.
%    noisemode  -  specifies whether the 'psnr' or 'sigma' fields should be
%                  used to compute the target sparse-coding error. this is
%                  useful when both fields are present in PARAMS. noisemode
%                  can either be 'psnr' or 'sigma'. if it is not specified,
%                  and both fields are present, psnr is used.
%
%
%  Optional fields in PARAMS (advanced):
%    gain       -  the noise gain. a positive value (usually close to 1)
%                  controlling the target error when sparse coding each
%                  block. when gain=1, the target error is precisely the
%                  value derived from psnr/sigma fields. when gain is
%                  different than 1, the target error is multiplied by this
%                  value. default is gain = 1.15.
%    lambda     -  the relative weight given to the input (noisy) signal in
%                  determining the denoised output. for each output sample,
%                  covered by K overlapping blocks, its value is computed
%                  by weighing the K denoised samples against the original
%                  noisy sample. in the averaging process, each denoised
%                  sample is attributed a relative weight of 1, and the
%                  noisy sample is attributed a relative weight of lambda.
%                  this way, samples with less confidence (i.e. covered by
%                  less blocks) are regularized by giving larger weight to
%                  the original sample. default value is maxval/(10*sigma),
%                  where sigma is the standard deviation of the noise.
%    maxatoms   -  specifies the limit on the number of atoms used to code
%                  each signal block (see parameter 'maxatoms' in
%                  OMP2/NNOMP2). default value is prod(blocksize)/2, i.e.
%                  half the number of samples in a block.
%    memusage   -  this parameter can be used to control the memory
%                  requirements of the function. memusage can be either
%                  'high', 'normal' (default value) or 'low'. when set to
%                  'low', the matrix G=D'*D is not precomputed before
%                  calling OMP, thus reducing memory requirements when the
%                  dictionary is highly redundant. 'high' and 'normal' are
%                  equivalent settings. note that setting memusage to 'low'
%                  will increase runtime substantially.
%
%
%  References:
%  [1] M. Elad and M. Aharon, "Image Denoising via Sparse and Redundant
%      representations over Learned Dictionaries", the IEEE Trans. on Image
%      Processing, Vol. 15, no. 12, pp. 3736-3745, December 2006.
%
%  See also OMPDENOISE2, OMPDENOISE3, KSVDDENOISE, OMP, OMP2.


%  Ron Rubinstein
%  Computer Science Department
%  Technion, Haifa 32000 Israel
%  ronrubin@cs
%
%  November 2007


% parse input arguments %

x = params.x;
p = ndims(x);
D = params.D;
blocksize = params.blocksize;


% blocksize %
if (numel(blocksize)==1)
  blocksize = ones(1,p)*blocksize;
end


% maxval %
if (isfield(params,'maxval'))
  maxval = params.maxval;
else
  maxval = 1;
end


% gain %
if (isfield(params,'gain'))
  gain = params.gain;
else
  gain = 1.15;
end


% maxatoms %
if (isfield(params,'maxatoms'))
  maxatoms = params.maxatoms;
else
  maxatoms = floor(prod(blocksize)/2);
end


% stepsize %
if (isfield(params,'stepsize'))
  stepsize = params.stepsize;
  if (numel(stepsize)==1)
    stepsize = ones(1,p)*stepsize;
  end
else
  stepsize = ones(1,p);
end
if (any(stepsize<1))
  error('Invalid step size.');
end


% noise mode %
if (isfield(params,'noisemode'))
  switch lower(params.noisemode)
    case 'psnr'
      sigma = maxval / 10^(params.psnr/20);
    case 'sigma'
      sigma = params.sigma;
    otherwise
      error('Invalid noise mode specified');
  end
elseif (isfield(params,'sigma'))
  sigma = params.sigma;
elseif (isfield(params,'psnr'))
  sigma = maxval / 10^(params.psnr/20);
else
  error('Noise strength not specified');
end


% lambda %
if (isfield(params,'lambda'))
  lambda = params.lambda;
else
  lambda = maxval/(10*sigma);
end


epsilon = sqrt(prod(blocksize)) * sigma * gain;   % target error for omp


MEM_LOW = 1;
MEM_NORMAL = 2;
MEM_HIGH = 3;

if (isfield(params,'memusage'))
  switch lower(params.memusage)
    case 'low'
      memusage = MEM_LOW;
    case 'normal'
      memusage = MEM_NORMAL;
    case 'high'
      memusage = MEM_HIGH;
    otherwise
      error('Invalid memory usage mode');
  end
else
  memusage = MEM_NORMAL;
end


% denoise the signal %

if (memusage >= MEM_NORMAL)
  G = D'*D;
else
  G = [];
end

y = zeros(size(x),class(x));

% ids{} contains the indices of the current block
ids = cell(p,1);
for j = 1:p
  ids{j} = 1:blocksize(j);
end

% lastids contains the indices of the last block in each dimension
lastids = stepsize .* floor((size(x)-blocksize)./stepsize) + 1;

blocknum = prod(floor((size(x)-blocksize)./stepsize) + 1);
tid = timerinit('ompdenoise', blocknum);

nz = 0;  % count non-zeros in block representations

for i = 1 : blocknum
 
  block = x(ids{:});
  block = block(:);
  dc = mean(block);
  block = block-dc;
  gamma = omp2(D'*block,sum(block.*block),G,epsilon,'maxatoms',maxatoms);
  nz = nz + nnz(gamma);
  y(ids{:}) = y(ids{:}) + reshape(D*gamma,blocksize) + dc;

  % increment block ids
  if (i<blocknum)
    j = 1;
    while (ids{j}(1) == lastids(j))
      ids{j} = 1:blocksize(j);
      j = j+1;
    end
    ids{j} = ids{j}+stepsize(j);
  end
  
  % display status
  if (nargin>=2 && msgdelta>0)
    timereta(tid, i, msgdelta);
  end

end

if (nargin>=2 && msgdelta>0)
  timereta(tid, i);
end

nz = nz/blocknum;  % average number of non-zeros

% average the denoised and noisy signals
cnt = countcover(size(x),blocksize,stepsize);
y = (y+lambda*x)./(cnt + lambda);

